"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
// builtin
const node_child_process_1 = require("node:child_process");
const node_process_1 = require("node:process");
// external
const json_1 = require("@bevry/json");
const fs_readable_1 = require("@bevry/fs-readable");
const fs_write_1 = __importDefault(require("@bevry/fs-write"));
// local
const index_js_1 = require("./index.js");
// cli
const argument_1 = __importDefault(require("@bevry/argument"));
const help = `
	ABOUT:
	Fetch backers (authors, maintainers, contributors, funders, sponsors, donors) and if desired output/write to package.json, json, string, text, markdown, html.

	USAGE:
	github-backers [...options] [-- [...options]]

	OPTIONS:
	--package=<string>
	  The package.json file to retrieve the data for and from.

	--slug=<string>
	  The GitHub Repository slug, defaults to fetching it from <package> or git.
	  This option is not reset with --.

	--[no-]offline[=<boolean>]
	  If provided, skip remote update, and only use the data from <package>

	--githubSponsorsUsername=<string>
	  Instead of autodetecting, use this username for fetching backers from GitHub Sponsors

	--opencollectiveUsername=<string>
	  Instead of autodetecting, use this username for fetching backers from OpenCollective

	--thanksdevGithubUsername=<string>
	  Instead of autodetecting, use this username for fetching backers from ThanksDev

	--sponsorCentsThreshold=<number>
	  The minimum amount of monthly cents to be considered a financial sponsor. Defaults to 100.

	--donorCentsThreshold=<number>
	  The minimum amount of eternal cents to be considered a financial donor. Defaults to 100.

	--
	  Act upon the previously specified options, and reset the following options:

	--write=<string>
	  The path that should be updated. If no path is provided, stdout is used.

	--format=<string>
	  The format of the file that should be updated, defaults to autodetection
	  Valid formats: string, text, markdown, html, package, copyright, shoutout, release, update`;
let result = null;
let renderOptions = {
    writePath: false,
};
const queryOptions = {
    githubSlug: null,
    packagePath: null,
    packageData: null,
    offline: false,
    githubSponsorsUsername: null,
    opencollectiveUsername: null,
    thanksdevGithubUsername: null,
    sponsorCentsThreshold: null,
    donorCentsThreshold: null,
};
function auto(value) {
    return value == null || value === true;
}
async function action() {
    // query
    if (auto(queryOptions.githubSlug)) {
        queryOptions.githubSlug =
            (0, index_js_1.getGitHubSlugFromUrl)((0, node_child_process_1.execSync)('git remote get-url origin').toString()) ||
                null;
    }
    if (auto(queryOptions.packagePath)) {
        if (await (0, fs_readable_1.isReadable)('package.json')) {
            queryOptions.packagePath = 'package.json';
            queryOptions.packageData = (await (0, json_1.readJSON)(queryOptions.packagePath));
        }
        else if (queryOptions.githubSlug &&
            typeof queryOptions.githubSlug === 'string') {
            queryOptions.packageData = await (0, index_js_1.getPackageData)(queryOptions.githubSlug);
            queryOptions.packagePath = 'package.json';
        }
    }
    if (auto(queryOptions.githubSlug) &&
        queryOptions.packageData &&
        typeof queryOptions.packageData === 'object') {
        queryOptions.githubSlug = (0, index_js_1.getGitHubSlugFromPackageData)(queryOptions.packageData);
    }
    // fetch
    if (!result)
        result = await (0, index_js_1.getBackers)(queryOptions);
    // write
    if (auto(renderOptions.writePath)) {
        renderOptions.writePath = queryOptions.packagePath || 'package.json';
    }
    if (queryOptions.packageData &&
        typeof queryOptions.packageData === 'object') {
        renderOptions.packageData = queryOptions.packageData;
    }
    if (auto(renderOptions.format)) {
        if (typeof renderOptions.writePath === 'string') {
            if (renderOptions.writePath.endsWith('package.json'))
                renderOptions.format = index_js_1.BackersRenderFormat.package;
            else if (renderOptions.writePath.endsWith('shoutout.txt') ||
                renderOptions.writePath.endsWith('shoutouts.txt'))
                renderOptions.format = index_js_1.BackersRenderFormat.shoutout;
            else if (renderOptions.writePath.endsWith('.json'))
                renderOptions.format = index_js_1.BackersRenderFormat.string;
            else if (renderOptions.writePath.endsWith('.txt'))
                renderOptions.format = index_js_1.BackersRenderFormat.text;
            else if (renderOptions.writePath.endsWith('.md'))
                renderOptions.format = index_js_1.BackersRenderFormat.markdown;
            else if (renderOptions.writePath.endsWith('.html'))
                renderOptions.format = index_js_1.BackersRenderFormat.html;
        }
        if (auto(renderOptions.format))
            renderOptions.format = index_js_1.BackersRenderFormat.string; // @todo add a raw mode
    }
    // output
    const output = (0, index_js_1.renderBackers)(result, renderOptions);
    if (typeof renderOptions.writePath === 'string') {
        if (renderOptions.writePath.endsWith('.json')) {
            if (typeof output === 'string') {
                console.warn(`Writing to ${renderOptions.writePath} as string... this is probably not intended, and you'll have to do some post-processing...`);
                await (0, fs_write_1.default)(renderOptions.writePath, output);
            }
            else {
                await (0, json_1.writeJSON)(renderOptions.writePath, output);
            }
        }
        else if (typeof output === 'string') {
            await (0, fs_write_1.default)(renderOptions.writePath, output);
        }
        else {
            console.warn(`Writing to ${renderOptions.writePath} as JSON... this is probably not intended, and you'll have to do some post-processing...`);
            await (0, json_1.writeJSON)(renderOptions.writePath, output);
        }
        console.log(`Wrote to ${renderOptions.writePath}`);
    }
    else {
        console.log(output);
    }
}
async function parse(args) {
    for (const arg of args) {
        const a = new argument_1.default(arg);
        switch (a.key) {
            case '--': {
                await action();
                renderOptions = {
                    writePath: false,
                };
                break;
            }
            case 'help': {
                return argument_1.default.help(help);
            }
            case 'slug': {
                result = null;
                queryOptions.packageData = null;
                queryOptions.githubSlug = a.string({ enabled: true, disabled: false });
                break;
            }
            case 'package': {
                result = null;
                queryOptions.packageData = null;
                queryOptions.packagePath = a.string({ enabled: true, disabled: false });
                break;
            }
            case 'offline': {
                queryOptions.offline = a.boolean();
                break;
            }
            case 'githubSponsorsUsername':
            case 'opencollectiveUsername':
            case 'thanksdevGithubUsername': {
                queryOptions[a.key] = a.string({ enabled: true, disabled: false });
                break;
            }
            case 'sponsorCentsThreshold':
            case 'donorCentsThreshold': {
                queryOptions[a.key] = a.number({
                    enabled: null,
                    disabled: 0,
                });
                break;
            }
            case 'write': {
                renderOptions.writePath = a.string({ enabled: true, disabled: false });
                break;
            }
            case 'format': {
                renderOptions.format = a.string({ enabled: '' });
                break;
            }
            default: {
                return a.unknown();
            }
        }
    }
    await action();
}
parse(node_process_1.argv.slice(2)).catch(argument_1.default.catch(help));
