import type { StrictUnion } from 'simplytyped';
import PromisePool from 'native-promise-pool';
import Errlop from 'errlop';
import Fellow from 'fellow';
export { Fellow, PromisePool, Errlop };
/** Field names that we consider backers */
export type BackerFields = 'author' | 'authors' | 'maintainers' | 'contributors' | 'funders' | 'sponsors' | 'donors';
/** Field names that we consider backers */
export declare const backerFields: Array<BackerFields>;
/** Collection of Sponsors, Funders, and Backers */
export interface Backers {
    author: Array<Fellow>;
    authors: Array<Fellow>;
    maintainers: Array<Fellow>;
    contributors: Array<Fellow>;
    funders: Array<Fellow>;
    sponsors: Array<Fellow>;
    donors: Array<Fellow>;
}
/** A rendering style for {@link Backers} */
export declare enum BackersRenderFormat {
    /** Renders as {@link BackersRenderResult.string} with appropriate defaults, intended for `package.json` merging. */
    string = "string",
    /** Renders as {@link BackersRenderResult.text} with appropriate defaults, intended for human-readable outputs.*/
    text = "text",
    /** Renders as {@link BackersRenderResult.markdown} with appropriate defaults, intended for readme outputs. */
    markdown = "markdown",
    /** Renders as {@link BackersRenderResult.html} with appropriate defaults, intended for readme outputs. */
    html = "html",
    /** Same as {@link BackersRenderFormat.string} however merges with {@link BackersRenderOptions.packageData} if it exists. */
    package = "package",
    /** Used for license files, author and authors are an array of strings, using {@link Fellow.toMarkdown} with appropriate defaults. */
    copyright = "copyright",
    /** Used for shoutout.txt files, outputs a string with shoutouts to contributors, funders, and sponsors, using {@link Fellow.toText} with appropriate defaults. */
    shoutout = "shoutout",
    /** Used for initial changelog entry, outputs a string with shoutout to funders and sponsors, using {@link Fellow.toMarkdown} with appropriate defaults. */
    release = "release",
    /** Used for subsequent changelog entries, outputs a string with shoutout to sponsors, using {@link Fellow.toMarkdown} with appropriate defaults. */
    update = "update"
}
/** Backers Render Result */
export interface BackersRenderResult {
    /** Renders using {@link Fellow.toString} with appropriate defaults, intended for `package.json` merging. */
    [BackersRenderFormat.string]: {
        author?: string;
        authors?: Array<string>;
        maintainers?: Array<string>;
        contributors?: Array<string>;
        funders?: Array<string>;
        sponsors?: Array<string>;
        donors?: Array<string>;
    };
    /** Renders using {@link Fellow.toText} with appropriate defaults, intended for human-readable outputs. */
    [BackersRenderFormat.text]: {
        author?: Array<string>;
        authors?: Array<string>;
        maintainers?: Array<string>;
        contributors?: Array<string>;
        funders?: Array<string>;
        sponsors?: Array<string>;
        donors?: Array<string>;
    };
    /** Renders using {@link Fellow.toMarkdown} with appropriate defaults, intended for readme outputs. */
    [BackersRenderFormat.markdown]: {
        author?: Array<string>;
        authors?: Array<string>;
        maintainers?: Array<string>;
        contributors?: Array<string>;
        funders?: Array<string>;
        sponsors?: Array<string>;
        donors?: Array<string>;
    };
    /** Renders using {@link Fellow.toHtml} with appropriate defaults, intended for readme outputs. */
    [BackersRenderFormat.html]: {
        author?: Array<string>;
        authors?: Array<string>;
        maintainers?: Array<string>;
        contributors?: Array<string>;
        funders?: Array<string>;
        sponsors?: Array<string>;
        donors?: Array<string>;
    };
    /** Same as {@link BackersRenderResult.string}, however merges with {@link BackersRenderOptions.packageData} if it exists. */
    [BackersRenderFormat.package]: {
        /** {@link BackersRenderOptions.packageData} properties*/
        [key: string]: any;
        author?: string;
        authors?: Array<string>;
        maintainers?: Array<string>;
        contributors?: Array<string>;
        funders?: Array<string>;
        sponsors?: Array<string>;
        donors?: Array<string>;
    };
    /** Renders using {@link Fellow.toHtml} with appropriate defaults, intended for readme outputs. */
    [BackersRenderFormat.copyright]: {
        /** Active copyright owners of the project, including copyright and years */
        author?: Array<string>;
        /** Eternal authors of the project, including copyright and years */
        authors?: Array<string>;
    };
    /** Renders using {@link Fellow.toText} with shoutouts for contributors, funders, and sponsors. */
    [BackersRenderFormat.shoutout]: string;
    /** Renders using {@link Fellow.toMarkdown} with shoutouts for funders and sponsors. */
    [BackersRenderFormat.release]: string;
    /** Renders using {@link Fellow.toMarkdown} with shoutouts for sponsors. */
    [BackersRenderFormat.update]: string;
}
/** Options for rendering {@link Backers} */
export interface BackersRenderOptions {
    /** The render format, if empty will throw. */
    format?: BackersRenderFormat | null;
    /** If provided with {@link BackersRenderFormat.package} will merge with the result data. */
    packageData?: PackageData | null;
    /** If provided with {@link BackersRenderFormat.html} will be used to generate contribtuion links. If `true` fills in from {@link BackersRenderOptions.packageData} if present. */
    githubSlug?: string | boolean | null;
    /** If provided with {@link BackersRenderFormat.shoutout} will be used to prefix each shoutout. If `true` fills in from {@link BackersRenderOptions.packageData} if present. */
    projectName?: string | boolean | null;
}
/** Options for queries. */
export interface QueryOptions {
    /** Set to non-zero if you wish to restrict how many requests are made at once. */
    concurrency?: number;
    /** The {@link PromisePool} instance to handle the {@link QueryOptions.concurrency} if specified. */
    pool?: PromisePool<any>;
    /** For REST and GraphQL APIs, the GitHub API Credentials */
    credentials?: GitHubCredentials;
    /** For GraphQL API, fetch this result field */
    resultField?: string;
    /** For GraphQL API, start results after this cursor. */
    afterCursor?: string;
    /** For REST and GraphQL APIs, how much results to return per page. */
    size?: number;
    /** For REST API, start results at this page. */
    page?: number;
    /** For REST API, set to non-zero if you wish to restrict how many pages are fetched. */
    pages?: number;
    /** For REST API, the URL to use on the fetch. */
    url?: string;
    /** For REST API, the pathname to use on the fetch. */
    pathname?: string;
    /** For REST API, the search/query params to use on the fetch. */
    searchParams?: URLSearchParams | Record<string, string>;
    /** For REST API, the headers to use on the fetch. */
    headers?: Record<string, string>;
    /** For REST API, the user agent to use on the fetch. */
    userAgent?: string;
    /** For REST API, the method to use on the fetch. */
    method?: RequestInit['method'];
    /** For REST API, the body to use on the fetch. */
    body?: RequestInit['body'];
}
/** Export some types we consume, so that others can also use them. Needs at least a {@link BackersQueryOptions.githubSlug} or {@link BackersQueryOptions.packageData} */
export interface BackersQueryOptions extends QueryOptions {
    /** The GitHub slug to determine backers for. If false, don't autodetect. */
    githubSlug?: string | boolean | null;
    /** The package.json data to determine backers for. If false, don't autodetect. */
    packageData?: PackageData | boolean | null;
    /** The GitHub Sponsors username to determine backers for. If false, don't autodetect. */
    githubSponsorsUsername?: string | boolean | null;
    /** The OpenCollective username to determine backers for. If false, don't autodetect. */
    opencollectiveUsername?: string | boolean | null;
    /** The ThanksDev GitHub username to determine backers for. If false, don't autodetect. */
    thanksdevGithubUsername?: string | boolean | null;
    /** If true, do not fetch any remote data when determing backers. */
    offline?: boolean | null;
    /** The minimum amount of monthly cents to be considered a financial donor (only applies if we are aware of the financial amount, defaults to 100). */
    sponsorCentsThreshold?: number | null;
    /** The minimum amount of eternal cents to be considered a financial donor (only applies if we are aware of the financial amount, defaults to 100). */
    donorCentsThreshold?: number | null;
}
/** Package.json person fields that we use */
export interface PackagePerson {
    name?: string;
    email?: string;
    username?: string;
    web?: string;
}
/** Package.json fields that we use */
export interface PackageData {
    /** The title of the project, used for rendering the default prefix */
    title?: string;
    /** The name of the project, used for rendering the default prefix if {@link PackageData.title} is not present. */
    name?: string;
    /** The homepage URL of the project */
    homepage?: string;
    /** The Repository URL of the project */
    repository?: string | {
        url?: string;
    };
    author?: string;
    authors?: Array<string>;
    maintainers?: Array<string>;
    contributors?: Array<string>;
    funders?: Array<string>;
    sponsors?: Array<string>;
    donors?: Array<string>;
    /** Badges configuration, see https://github.com/bevry/badges and https://github.com/bevry/projectz */
    badges?: {
        config?: {
            githubSponsorsUsername?: string;
            opencollectiveUsername?: string;
            thanksdevGithubUsername?: string;
        };
    };
}
/** Fetch the package data of a repository. */
export declare function getPackageData(githubSlug: string, fallback?: PackageData): Promise<PackageData>;
/**
 * Get the GitHub slug of this GitHub url
 * @example
 * ```
 * bevry/github-api
 * github:bevry/github-api
 * git@github.com:bevry/github-api.git
 * https://github.com/bevry/github-api.git
 * ssh://github.com/bevry/github-api.git
 * git://github.com/user/project.git#commit-ish
 * git+https://github.com/bevry/github-api.git#commit-ish
 * git+ssh://github.com/bevry/github-api.git#commit-ish
 * git+ssh+ssh://github.com/bevry/github-api.git#commit-ish
 * ```
 */
export declare function getGitHubSlugFromUrl(url: string): string | null;
/** Get the GitHub Repository website url of a GitHub slug */
export declare function getRepositoryWebsiteUrlFromGitHubSlugOrUrl(githubSlug?: string): string | null;
/** Get the GitHub Repository issues url of a GitHub slug */
export declare function getRepositoryIssuesUrlFromGitHubSlugOrUrl(githubSlug?: string): string | null;
/** Get the GitHub Repository url of a GitHub slug or GitHub url */
export declare function getRepositoryUrlFromGitHubSlugOrUrl(githubSlug: string): string | null;
/** Get the Repository url of a URL or GitHub slug */
export declare function getRepositoryUrlFromUrlOrGitHubSlug(input: string): string | null;
/** Get the Repository url (or shorthand url) from package.json data */
export declare function getRepositoryUrlFromPackageData(packageData: Pick<PackageData, 'homepage' | 'repository'>): string | null;
/** Get the GitHub slug from package.json data  */
export declare function getGitHubSlugFromPackageData(packageData: Pick<PackageData, 'homepage' | 'repository'>): string | null;
/** A .github/FUNDING.yml response */
export interface FundingData {
    github?: string | Array<string>;
    patreon?: string | Array<string>;
    open_collective?: string | Array<string>;
    ko_fi?: string | Array<string>;
    liberapay?: string | Array<string>;
    custom?: string | Array<string>;
}
export declare enum ThanksDevPlatform {
    GitHub = "gh",
    GitLab = "gl"
}
export type ThanksDevDonor = [
    platform: ThanksDevPlatform,
    username: string,
    /** Can be zero */
    cents: number
];
export interface ThanksDevResponse {
    donors: Array<ThanksDevDonor>;
    dependers: Array<ThanksDevDonor>;
}
export interface ThanksDevBackers {
    sponsors: Array<Fellow>;
    donors: Array<Fellow>;
}
/** Get the {@link Backers} of the from the ThanksDev API */
export declare function getBackersFromThanksDev(platform: ThanksDevPlatform, username: string, opts?: BackersQueryOptions): Promise<ThanksDevBackers>;
export interface OpenCollectiveMember {
    MemberId: number;
    createdAt: string;
    type: 'USER' | 'ORGANIZATION';
    role: 'ADMIN' | 'HOST' | 'BACKER';
    isActive: boolean;
    /** Dollars */
    totalAmountDonated: number;
    currency: string;
    lastTransactionAt: string;
    /** Dollars */
    lastTransactionAmount: number;
    profile: string;
    name: string;
    company: null | string;
    description: null | string;
    image: null | string;
    email?: null;
    twitter: null | string;
    github: null | string;
    website: null | string;
}
export type OpenCollectiveResponse = Array<OpenCollectiveMember>;
export interface OpenCollectiveBackers {
    sponsors: Array<Fellow>;
    donors: Array<Fellow>;
}
/** Get the {@link Backers} of the from the OpenCollective API */
export declare function getBackersFromOpenCollective(username: string, opts?: BackersQueryOptions): Promise<OpenCollectiveBackers>;
/** GitHub's response when an error occurs. */
export interface GitHubError {
    message?: string;
    documentation_url?: string;
    errors?: Array<{
        type: string;
        path: Array<string>;
        locations: Array<{
            line: number;
            column: number;
        }>;
        message: string;
    } | {
        resource: string;
        field: string;
        code: string;
    }>;
}
/** If the variable `GITHUB_API_URL` or `GITHUB_API` exists, use that, otherwise use the value `https://api.github.com`. */
export type GitHubApiUrl = StrictUnion<{
    /** https://docs.github.com/en/free-pro-team@latest/actions/reference/environment-variables#default-environment-variables */
    GITHUB_API_URL: string;
} | {
    /** https://github.com/search?l=JavaScript&q=GITHUB_API&type=Code */
    GITHUB_API: string;
}>;
/**
 * If the variable `GITHUB_ACCESS_TOKEN` or `GITHUB_TOKEN` exists, use that according to:
 * https://developer.github.com/v3/#oauth2-token-sent-as-a-parameter
 */
export type GitHubToken = StrictUnion<{
    /** https://github.com/search?q=GITHUB_ACCESS_TOKEN&type=code */
    GITHUB_ACCESS_TOKEN: string;
} | {
    /** https://docs.github.com/en/actions/reference/authentication-in-a-workflow#about-the-github_token-secret */
    GITHUB_TOKEN: string;
}>;
/**
 * If the variables `GITHUB_CLIENT_ID` and `GITHUB_CLIENT_SECRET` exist, use that according to:
 * https://developer.github.com/v3/#oauth2-keysecret
 */
export interface GitHubClient {
    GITHUB_CLIENT_ID: string;
    GITHUB_CLIENT_SECRET: string;
}
/**
 * Provide `GITHUB_ACCESS_TOKEN` or `GITHUB_TOKEN`,
 * or a combination of `GITHUB_CLIENT_ID` and `GITHUB_CLIENT_SECRET`,
 * optionally with `GITHUB_ACCESS_TOKEN` or `GITHUB_TOKEN`.
 * https://developer.github.com/v3/#authentication
 */
export type GitHubCredentials = StrictUnion<GitHubToken | GitHubClient> & Partial<GitHubApiUrl>;
export type GitHubGraphQLResponse<T> = StrictUnion<GitHubError | T>;
export type GitHubRESTResponse<T> = StrictUnion<GitHubError | T>;
export interface GitHubUserGraphQL {
    bio: string;
    company: string;
    email: string;
    isHireable: boolean;
    location: string;
    login: string;
    name: string;
    url: string;
    websiteUrl: string;
}
export interface GitHubOrganizationGraphQL {
    description: string;
    email: string;
    location: string;
    login: string;
    name: string;
    url: string;
    websiteUrl: string;
}
export interface GitHubProfileGraphQL {
    user?: GitHubUserGraphQL;
    organization?: GitHubOrganizationGraphQL;
}
export type GitHubSponsorGraphQL = GitHubUserGraphQL & GitHubOrganizationGraphQL;
export interface GitHubSponsorsGraphQL {
    user: {
        sponsors: {
            pageInfo?: {
                hasNextPage: boolean;
                endCursor: string;
            };
            nodes: Array<GitHubSponsorGraphQL>;
        };
    };
}
export interface GitHubSponsorsBackers {
    sponsors: Array<Fellow>;
    donors: Array<Fellow>;
}
export interface GitHubCommitREST {
    sha: string;
}
/**
 * GitHub's response to getting a user.
 * https://developer.github.com/v3/users/#get-a-single-user
 */
export interface GitHubProfileREST {
    login: string;
    id: number;
    node_id: string;
    avatar_url: string;
    gravatar_id: string;
    url: string;
    html_url: string;
    followers_url: string;
    following_url: string;
    gists_url: string;
    starred_url: string;
    subscriptions_url: string;
    organizations_url: string;
    repos_url: string;
    events_url: string;
    received_events_url: string;
    type: string;
    site_admin: boolean;
    name: string;
    company: string;
    blog: string;
    location: string;
    email: string;
    hireable: boolean;
    bio: string;
    public_repos: number;
    public_gists: number;
    followers: number;
    following: number;
    created_at: string;
    updated_at: string;
}
/** The organization of a repository */
export interface GitHubOrganizationREST {
    login: string;
    id: number;
    node_id: string;
    avatar_url: string;
    gravatar_id: string;
    url: string;
    html_url: string;
    followers_url: string;
    following_url: string;
    gists_url: string;
    starred_url: string;
    subscriptions_url: string;
    organizations_url: string;
    repos_url: string;
    events_url: string;
    received_events_url: string;
    type: string;
    site_admin: boolean;
}
/** The permissions of a repository */
export interface GitHubPermissionsREST {
    admin: boolean;
    push: boolean;
    pull: boolean;
}
/**
 * GitHub's response to getting a repository.
 * https://developer.github.com/v3/repos/#list-contributors
 */
export interface GitHubContributorREST {
    login: string;
    id: number;
    node_id: string;
    avatar_url: string;
    gravatar_id: string;
    url: string;
    html_url: string;
    followers_url: string;
    following_url: string;
    gists_url: string;
    starred_url: string;
    subscriptions_url: string;
    organizations_url: string;
    repos_url: string;
    events_url: string;
    received_events_url: string;
    type: string;
    site_admin: boolean;
    /** Count of how many contributions */
    contributions: number;
}
export type GitHubContributorsREST = Array<GitHubContributorREST>;
/**
 * GitHub's response to getting a members of an organization
 * https://developer.github.com/v3/orgs/members/#members-list
 */
export interface GitHubMemberREST {
    login: string;
    id: number;
    node_id: string;
    avatar_url: string;
    gravatar_id: string;
    url: string;
    html_url: string;
    followers_url: string;
    following_url: string;
    gists_url: string;
    starred_url: string;
    subscriptions_url: string;
    organizations_url: string;
    repos_url: string;
    events_url: string;
    received_events_url: string;
    type: string;
    site_admin: boolean;
}
export type GitHubMembersREST = Array<GitHubMemberREST>;
/**
 * GitHub's response to searching for repositories
 * https://developer.github.com/v3/search/#search-repositories
 */
export interface GitHubSearchREST {
    total_count: number;
    incomplete_results: boolean;
    items: GitHubSearchRepositoryREST[];
}
export type GitHubsearchResponseREST = StrictUnion<GitHubError | GitHubSearchREST>;
/** The license of a repository */
export interface GitHubLicenseREST {
    key: string;
    name: string;
    spdx_id: string;
    url: string;
    node_id: string;
}
/** The owner of a repository */
export interface GitHubOwnerREST {
    login: string;
    id: number;
    node_id: string;
    avatar_url: string;
    gravatar_id: string;
    url: string;
    received_events_url: string;
    type: string;
}
/** Search results return a subset of the full repository results */
export interface GitHubSearchRepositoryREST {
    id: number;
    node_id: string;
    name: string;
    full_name: string;
    owner: GitHubOwnerREST;
    private: boolean;
    html_url: string;
    description: string;
    fork: boolean;
    url: string;
    created_at: Date;
    updated_at: Date;
    pushed_at: Date;
    homepage: string;
    size: number;
    stargazers_count: number;
    watchers_count: number;
    language: string;
    forks_count: number;
    open_issues_count: number;
    master_branch: string;
    default_branch: string;
    score: number;
}
/**
 * GitHub's response to getting a repository
 * https://developer.github.com/v3/repos/#get
 */
export interface GitHubRepositoryREST extends GitHubSearchRepositoryREST {
    archive_url: string;
    assignees_url: string;
    blobs_url: string;
    branches_url: string;
    collaborators_url: string;
    comments_url: string;
    commits_url: string;
    compare_url: string;
    contents_url: string;
    contributors_url: string;
    deployments_url: string;
    downloads_url: string;
    events_url: string;
    forks_url: string;
    git_commits_url: string;
    git_refs_url: string;
    git_tags_url: string;
    git_url: string;
    issue_comment_url: string;
    issue_events_url: string;
    issues_url: string;
    keys_url: string;
    labels_url: string;
    languages_url: string;
    merges_url: string;
    milestones_url: string;
    notifications_url: string;
    pulls_url: string;
    releases_url: string;
    ssh_url: string;
    stargazers_url: string;
    statuses_url: string;
    subscribers_url: string;
    subscription_url: string;
    tags_url: string;
    teams_url: string;
    trees_url: string;
    clone_url: string;
    mirror_url: string;
    hooks_url: string;
    svn_url: string;
    is_template: boolean;
    topics: string[];
    has_issues: boolean;
    has_projects: boolean;
    has_wiki: boolean;
    has_pages: boolean;
    has_downloads: boolean;
    archived: boolean;
    disabled: boolean;
    visibility: string;
    permissions: GitHubPermissionsREST;
    allow_rebase_merge: boolean;
    template_repository: null;
    allow_squash_merge: boolean;
    allow_merge_commit: boolean;
    subscribers_count: number;
    network_count: number;
    license?: GitHubLicenseREST;
    organization?: GitHubOrganizationREST;
    parent?: GitHubRepositoryREST;
    source?: GitHubRepositoryREST;
}
/**
 * Get the desired GitHub Access Token from the credentials.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
export declare function getAccessToken(credentials?: GitHubCredentials): string | null;
/** Check whether or not sufficient GitHub credentials were supplied. */
export declare function hasCredentials(credentials?: GitHubCredentials): boolean;
/**
 * Check whether or not sufficient GitHub credentials were supplied.
 * @throws if invalid
 */
export declare function validateCredentials(credentials?: GitHubCredentials): true;
/**
 * Get the GitHub Authorization Search Params.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * @param credentials The params to use for the authorization variables.
 * @param params If you wish to set the params on an existing URLSearchParams instance, then provide it here.
 * @returns The search params that you should append to your github API request url.
 * @throws If no valid GitHub Authorization was provided.
 */
export declare function getSearchParams(credentials?: GitHubCredentials, params?: URLSearchParams): URLSearchParams;
export declare function applySearchParams(target: URLSearchParams, source: URLSearchParams | Record<string, string> | null | undefined): void;
/**
 * Remove any GitHub Credentials from a URL Search Params instance.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
export declare function removeSearchParams(params: URLSearchParams): URLSearchParams;
/**
 * Redact any GitHub Credentials in their URLSearchParams from a string.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * @param value The string to redact credentials from.
 * @returns The string with the credentials redacted.
 */
export declare function redactSearchParams(value: string): string;
/** Redact the GitHub Credentials in any form from a string */
export declare function redactCredentials(value: string, credentials?: GitHubCredentials): string;
/**
 * Get the GitHub Authorization as a Query String.
 * You probably want to use {@link getURL} directly, instead of going through this method.
 */
export declare function getQueryString(credentials?: GitHubCredentials): string;
/**
 * Get the GitHub Authorization Header.
 * Use as the `Authorization` header within {@link fetch} calls.
 * You probably want to use {@link getHeaders} or {@link fetch} directly, instead of going through this method.
 * @throws If no valid GitHub Authorization was provided.
 */
export declare function getAuthHeader(credentials?: GitHubCredentials): string;
/**
 * Get the headers to attach to the request to the GitHub API.
 * Use as the headers object within {@link fetch} calls.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
export declare function getHeaders(credentials?: GitHubCredentials, headers?: Record<string, string>): {
    Accept: string;
    Authorization: string;
};
/**
 * Remove any GitHub Credentials from a Headers instance.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 */
export declare function removeHeaders(headers: Record<string, string>): Record<string, string>;
/**
 * Get the desired Github API URL, using {@link removeSearchParams}.to ensure there are no credentials.
 * As this URL does not include credentials, use with {@link getAuthHeader} to authorize correctly.
 * Otherwise use {@link getCredentialedURL} to get a credentialed URL.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * If the credentials property is nullish, then the environment variables are attempted.
 */
export declare function getURL(opts?: Pick<QueryOptions, 'credentials' | 'url' | 'pathname' | 'searchParams'>): URL;
/**
 * Get the credentialed GitHub API URL instance.
 * Uses {@link getURL} to get the URL, then uses {@link getSearchParams} to add the credentials.
 * You probably want to use {@link fetch} directly, instead of going through this method.
 * If the credentials property is nullish, then the environment variables are attempted.
 */
export declare function getCredentialedURL(opts?: Pick<QueryOptions, 'credentials' | 'url' | 'pathname' | 'searchParams'>): URL;
/**
 * Fetches a GitHub REST API response with authentication, parsing, waiting, pooling, paging.
 * If the credentials property is nullish, then the environment variables are attempted.
 * If the user agent is nullish, then it will be set to `"@bevry/github"`
 */
export declare function queryREST<T>(opts?: QueryOptions): Promise<T>;
/**
 * Fetches a GitHub GraphQL API response with authentication, parsing, waiting, pooling. NO PAGING.
 * If the credentials property is nullish, then the environment variables are attempted.
 * If the user agent is nullish, then it will be set to `"@bevry/github"`
 */
export declare function queryGraphQL<T>(query: string, opts?: QueryOptions): Promise<T>;
/** Get the latest commit for a GitHub Repository */
export default function getGitHubLatestCommit(githubSlug: string, opts?: QueryOptions): Promise<string>;
/** Get this GitHub Repository */
export declare function getGitHubRepository(githubSlug: string, opts?: QueryOptions): Promise<GitHubRepositoryREST>;
/** Get the GitHub Repositories of these slugs */
export declare function getGitHubRepositories(githubSlugs: string[], opts?: QueryOptions): Promise<Array<GitHubRepositoryREST>>;
/**
 * Get the GitHub Repositories of this search query
 * @param search e.g. `\@bevry language:typescript`
 */
export declare function getGitHubRepositoriesFromSearch(search: string, opts?: QueryOptions): Promise<Array<GitHubSearchRepositoryREST>>;
/** Get the GitHub Repositories of these usernames (users, organizations) */
export declare function getGitHubRepositoriesFromUsernames(usernames: string[], opts?: QueryOptions): Promise<Array<GitHubSearchRepositoryREST>>;
/** Get the {@link Fellow} entity of this GitHub User via the GitHub GraphQL API */
export declare function getGitHubUser(username: string, opts?: BackersQueryOptions): Promise<Fellow>;
/** Get the {@link Fellow} entity of this GitHub Organization via the GitHub GraphQL API */
export declare function getGitHubOrganization(username: string, opts?: BackersQueryOptions): Promise<Fellow>;
/** Get the {@link Fellow} entity of this GitHub User/Organisation via the GitHub REST API */
export declare function getGitHubProfileFromApiUrl(url: string, opts?: QueryOptions): Promise<Fellow>;
/** Get the {@link Fellow} entity of this GitHub User/Organisation via the GitHub API */
export declare function getGitHubProfile(username: string, opts?: QueryOptions): Promise<Fellow>;
/** Get the {@link Fellow} entities for the GitHub Sponsors of this GitHub Profile via the GitHub Sponsors API */
export declare function getBackersFromGitHubSponsors(username: string, opts?: BackersQueryOptions): Promise<GitHubSponsorsBackers>;
/** Get the {@link Fellow} entities for GitHub Contributors (minus bots) of this GitHub Repository via the GitHub REST API */
export declare function getGitHubContributors(githubSlug: string, opts?: QueryOptions): Promise<Array<Fellow>>;
/** Get the {@link Fellow} entities for GitHub Members of this GitHub Organisation via the GitHub REST API */
export declare function getGitHubMembersFromOrganization(org: string, opts?: QueryOptions): Promise<Array<Fellow>>;
/** Get the {@link Fellow} entities for the GitHub Members of these GitHub Organisations via the GitHub REST API */
export declare function getGitHubMembersFromOrganizations(orgs: Array<string>, opts?: BackersQueryOptions): Promise<Array<Fellow>>;
/** Get the {@link Backers} of this GitHub Repository */
export declare function getBackers(opts?: BackersQueryOptions): Promise<Backers>;
/** Get the {@link Backers} of these GitHub Repositories */
export declare function getBackersFromRepositories(githubSlugs: Array<string>, opts?: BackersQueryOptions): Promise<Backers>;
/** Get the {@link Backers} for the GitHub Repositories of these usernames (users, organizations) */
export declare function getBackersFromUsernames(usernames: Array<string>, opts?: BackersQueryOptions): Promise<Backers>;
/**
 * Get the {@link Backers} for the GitHub Repositories of this search query
 * @param query e.g.`\@bevry language:typescript`
 */
export declare function getBackersFromSearch(query: string, opts?: BackersQueryOptions): Promise<Backers>;
export declare function renderBackers<T extends BackersRenderFormat>(backers: Partial<Backers>, opts: BackersRenderOptions & {
    format?: undefined | null;
}): never;
export declare function renderBackers<T extends BackersRenderFormat>(backers: Partial<Backers>, opts: BackersRenderOptions & {
    format: T;
}): BackersRenderResult[T];
//# sourceMappingURL=index.d.ts.map